import json
import uuid
from contextvars import ContextVar


class BasePoolManager:
    def __init__(self, client_context_manager):
        self._pools = {}
        self._config = {}
        self._ready = False
        self._client_context_manager = client_context_manager
        self._contexts = {}

    def register(self, config):
        assert not self.ready, f'{self.__class__.__name__} already registered'
        self._contexts = {
            pool_name: ContextVar(f'{self.__class__.__name__}_{pool_name}_{uuid.uuid1()}', default=None)
            for pool_name in config
        }
        self._config = config

    async def initialize(self):

        await self

    def __await__(self):

        raise NotImplementedError()

    def __repr__(self):
        return id(self)

    def __str__(self):
        return json.dumps({
            pool_name: self.pool_status(pool_name)
            for pool_name, pool in self._pools.items()
        }, indent=4, ensure_ascii=False)

    @property
    def ready(self):

        return self._ready

    def func_partial(self, conn, pool_name):

        raise NotImplementedError()

    def pool_status(self, pool_name='default'):

        raise NotImplementedError()

    def _echo_pool_info(self, pool_name='default'):

        raise NotImplementedError()

    def get_context_client(self, pool_name='default'):
        ref = self._contexts[pool_name].get()
        client = None
        if ref is not None:
            _client = ref() or None
            if _client and not _client.closed:
                client = _client
        return client

    def get_client(self, pool_name='default'):
        self._echo_pool_info(pool_name)
        _client = self.get_context_client(pool_name)
        if _client:
            return self._client_context_manager(_client, self.func_partial)

        return self._client_context_manager(self._pools[pool_name], self.func_partial)

    async def close(self):

        raise NotImplementedError()




import asyncio
import functools
import weakref
import aiomysql

from aiomysql.sa import SAConnection, Engine
from aiomysql.sa.engine import _dialect
from sqlalchemy.sql.selectable import Select
from sqlalchemy.sql.dml import Insert, Update, Delete
from wintersweet.asyncs.pool.base import BasePoolManager
from wintersweet.asyncs.pool.context import MysqlAsyncContextManager
from wintersweet.utils.base import Utils
from wintersweet.utils.metaclass import SafeSingleton


class MysqlClient:
    """MySQL客户端
    """
    def __init__(self, conn: SAConnection):
        self._conn = conn
        self._lock = asyncio.Lock()

    def __del__(self):
        if self._conn and not self._conn.closed:
            _conn, self._conn = self._conn, None
            Utils.asyncio.create_task(_conn.close)
            Utils.log.critical(f'Mysql:{self._conn} conn leakage!')

    @staticmethod
    def safestr(val):

        cls = type(val)

        if cls is str:
            val = aiomysql.escape_string(val)
        elif cls is dict:
            val = aiomysql.escape_dict(val, 'utf8')
        else:
            val = str(val)

        return val

    @property
    def conn(self):
        return self._conn

    @property
    def closed(self):
        return self._conn.closed

    def close(self):
        return self._conn.close()

    async def execute(self, query, *multiparams, **params):

        result = None

        async with self._lock:

            result = await self._conn.execute(query, *multiparams, **params)

        return result

    async def select(self, query, *multiparams, **params):

        result = []

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            records = await proxy.cursor.fetchall()

            if records:
                result = records

            if not proxy.closed:
                await proxy.close()

        return result

    async def find(self, query, *multiparams, **params):

        result = None

        if not isinstance(query, Select):
            raise TypeError(r'Not sqlalchemy.sql.selectable.Select object')

        proxy = await self.execute(query.limit(1), *multiparams, **params)

        if proxy is not None:

            record = await proxy.cursor.fetchone()

            if record:
                result = record

            if not proxy.closed:
                await proxy.close()

        return result

    async def count(self, query, *multiparams, **params):

        result = await self.find(query, *multiparams, **params)

        return result[r'tbl_row_count']

    async def insert(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Insert):
            raise TypeError(r'Not sqlalchemy.sql.dml.Insert object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = self._conn.connection.insert_id()

            if not proxy.closed:
                await proxy.close()

        return result

    async def update(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Update):
            raise TypeError(r'Not sqlalchemy.sql.dml.Update object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = proxy.rowcount

            if not proxy.closed:
                await proxy.close()

        return result

    async def delete(self, query, *multiparams, **params):

        result = 0

        if not isinstance(query, Delete):
            raise TypeError(r'Not sqlalchemy.sql.dml.Delete object')

        proxy = await self.execute(query, *multiparams, **params)

        if proxy is not None:

            result = proxy.rowcount

            if not proxy.closed:
                await proxy.close()

        return result

    def begin(self):
        return self._conn.begin()


class MysqlPoolManager(SafeSingleton, BasePoolManager):
    """数据库连接池管理器"""

    def __init__(self):

        super(MysqlPoolManager, self).__init__(MysqlAsyncContextManager)
        self._engines = {}

    def __await__(self):
        if self._ready:
            return self

        for pool_name, pool_config in self._config.items():
            pool = yield from aiomysql.create_pool(**pool_config).__await__()
            self._pools[pool_name] = pool
            setattr(self._pools[pool_name], 'pool_name', pool_name)

            self._engines[pool_name] = Engine(_dialect, pool)
            Utils.log.success(
                f"mysql pool {self.pool_status(pool_name)} initialized"
            )
        self._ready = True

        return self

    def func_partial(self, conn, pool_name):
        client = MysqlClient(SAConnection(conn, self._engines[pool_name]))
        self._contexts[pool_name].set(weakref.ref(client))

        return client

    def pool_status(self, pool_name='default'):
        pool = self._pools[pool_name]
        user = self._config[pool_name].get('user', 'root')
        host = self._config[pool_name].get('host', 'localhost')
        db = self._config[pool_name].get('db', None)
        port = self._config[pool_name].get('port', 3306)

        return f'<[{pool_name}] {user}@{host}:{port} [db:{db}, size:[{pool.minsize}:{pool.maxsize}], free:{pool.freesize}]>'

    def _echo_pool_info(self, pool_name='default'):

        assert self._ready, f'{self.__class__.__name__} does not initialize'
        pool = self._pools[pool_name]
        if (pool.maxsize - pool.size + pool.freesize) < Utils.math.ceil(pool.maxsize / 3):
            Utils.log.warning(
                f'Mysql pool not enough ({pool_name}): {pool.freesize}({pool.size}/{pool.maxsize})'
             )

    async def find(self, sql: Select, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.find(sql, *args, **kwargs)

    async def select(self, sql: Select, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.select(sql, *args, **kwargs)

    async def count(self, sql, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.count(sql, *args, **kwargs)

    async def update(self, sql: Update, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.update(sql, *args, **kwargs)

    async def insert(self, sql, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.insert(sql, *args, **kwargs)

    async def delete(self, sql: Delete, *args, pool_name='default', **kwargs):

        async with self.get_client(pool_name) as client:
            return await client.delete(sql, *args, **kwargs)

    async def execute(self, query, *multiparams, pool_name='default', **params):

        async with self.get_client(pool_name) as client:
            return await client.execute(query, *multiparams, **params)

    def close(self):
        self._ready = False
        _pools, self._pools = self._pools, None
        for pool in _pools.values():
            pool.close()


mysql_pool_manager = MysqlPoolManager()


def atomic(pool_name='default'):

    def wrapper(func):
        @functools.wraps(func)
        async def callback(*args, **kwargs):
            global mysql_pool_manager
            if not mysql_pool_manager.ready:
                await mysql_pool_manager

            async with mysql_pool_manager.get_client(pool_name) as trx:
                async with trx.begin():
                    result = await func(*args, **kwargs)
                    return result
        return callback
    return wrapper




import time
import traceback
import weakref
import aioredis

from aioredis import Redis, ConnectionsPool

from wintersweet.asyncs.pool.base import BasePoolManager
from wintersweet.asyncs.pool.context import RedisAsyncContextManager
from wintersweet.asyncs.task.tasks import IntervalTask
from wintersweet.asyncs.tools.circular import AsyncCircularForTimeout
from wintersweet.utils.base import Utils
from wintersweet.utils.errors import catch_error
from wintersweet.utils.metaclass import SafeSingleton


class MutexLock:
    """基于Redis实现的分布式锁，当await lock.acquire()一旦获取锁成功，会自动触发看门狗，持续对lock进行检查并保持持有
        直到跳出async with管理或者手动调用release()
        使用方法1（建议）：
            async with MutexLock(redis_pool, 'test-key', 60) as lock:
                is_locked = await lock.acquire()
                if is_locked:
                    # do something
                else:
                    # do something

        当你需要在程序生命周期内保持锁占有时，可使用方法2
        使用方法2：
            lock = MutexLock(redis_pool, 'test-key', 60)
            try:
                is_locked = await lock.acquire()
                if is_locked:
                    # do something
                else:
                    # do something
            except Exception as e:
                # do something
            finally:
                await lock.release()
    """

    _renew_script = '''
    if redis.call("get",KEYS[1]) == ARGV[1] and redis.call("ttl",KEYS[1]) > 0 then
        return redis.call("expire",KEYS[1],ARGV[2])
    else
        return 0
    end
    '''

    _unlock_script = '''
    if redis.call("get",KEYS[1]) == ARGV[1] then
        return redis.call("del",KEYS[1])
    else
        return 0
    end
    '''

    def __init__(self, redis_pool: ConnectionsPool, key, expire=60, commands_factory=Redis):

        self._redis_pool = redis_pool
        assert expire > 3, '"expire" is too small'
        self._expire = expire
        self._commands_factory = commands_factory
        self._key = key
        self._lock_tag = f'process_lock_{key}'
        self._lock_val = Utils.uuid.uuid1().encode()

        self._locked = False
        self._cache = None
        self._conn = None
        self._err_count = 0

        # watch dog
        self._watcher = None

    async def __aenter__(self):

        await self._init_conn()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):

        await self.release()

    def __del__(self):
        if self._watcher.running():
            self._watcher.stop()
            self._watcher = None

        if self._conn:
            self._redis_pool.release(self._conn)
            self._cache = self._conn = None
            self._redis_pool = None

    def __repr__(self):
        return f'{self.__class__.__name__}<{self._key}>'

    @property
    def locked(self):
        return self._locked

    async def _do_heartbeat(self):
        """watch dog continuous to refresh the lock"""
        res = await self.acquire(0)
        Utils.log.info(f'{str(self)} Watcher lock result: {res}')

    async def _init_conn(self, reset=False):
        """init redis conn or reset redis conn"""
        try:
            if self._conn is None or reset:
                if reset and self._conn is not None:
                    await self.release()
                self._conn = await self._redis_pool.acquire()
                self._cache = self._commands_factory(self._conn)
                self.init_watcher()
                return True
        except Exception:
            Utils.log.error(f'{str(self)} init redis conn error')
            Utils.log.error(traceback.format_exc())

    def init_watcher(self):
        """init watch dog"""
        interval = Utils.math.floor(self._expire / 3)
        self._watcher = IntervalTask(interval or 3, self._do_heartbeat, tag=self.__class__.__name__)

    async def exists(self):
        """check the lock exist or not"""
        await self._init_conn()
        if await self._cache.exists(self._lock_tag):
            return True
        else:
            return False

    async def acquire(self, timeout=0):
        """acquire the lock"""
        try:
            await self._init_conn()
            if self._locked:
                self._locked = await self.renew()
            else:
                params = {
                    r'key': self._lock_tag,
                    r'value': self._lock_val,
                    r'expire': self._expire,
                    r'exist': Redis.SET_IF_NOT_EXIST,
                }

                async for _ in AsyncCircularForTimeout(timeout):

                    if await self._cache.set(**params):
                        self._locked = True

                    if self._locked or timeout == 0:
                        break
            if self._locked and not self._watcher.running():
                self._watcher.start()

        except aioredis.errors.ConnectionClosedError:
            self._err_count += 1
            Utils.log.error(f'{str(self)} acquire conn closed error')
            if self._err_count >= 3:
                # If the connection is damaged, automatic reset to ensure service resume
                reset_res = await self._init_conn(reset=True)
                if reset_res:
                    self._err_count = 0

        return self._locked

    async def wait(self, timeout=0):

        async for _ in AsyncCircularForTimeout(timeout):

            if not await self.exists():
                return True
        else:
            return False

    async def renew(self):
        """renew the lock when lock held"""
        if self._cache is None:
            self._locked = False

        if self._locked:
            self._locked = False
            if await self._cache.eval(self._renew_script, [self._lock_tag], [self._lock_val, self._expire]):
                self._locked = True
            else:
                self._locked = False

        return self._locked

    async def release(self):
        """release the lock, and stop the watch dog"""
        with catch_error():
            if self._watcher.running():
                self._watcher.stop()

            if self._locked and self._cache:
                await self._cache.eval(self._unlock_script, [self._lock_tag], [self._lock_val])
                self._locked = False

            if self._conn:
                self._redis_pool.release(self._conn)
                self._cache = self._conn = None


class FrequencyLimiter:

    DEFAULT_EXPIRE = 60

    def __init__(self, redis_pool, unique_tag, granularity, limit, ntp=None):

        self._unique_tag = unique_tag
        self._redis_pool = redis_pool
        self._granularity = granularity
        self._limit = limit
        self._ntp = ntp

    def generate_key(self):

        if self._ntp:
            divide = int(self._ntp.timestamp) // self._granularity
        else:
            divide = int(time.time()) // self._granularity

        return f'{self.__class__.__name__}_{self._unique_tag}_{divide}'

    async def incrby(self, incr_num=1):

        pipeline = Redis(self._redis_pool).pipeline()
        key = self.generate_key()

        pipeline.incrby(key, incr_num)
        pipeline.expire(key, max([self._granularity, self.DEFAULT_EXPIRE]))

        res, _ = await pipeline.execute()
        return res

    async def is_limited(self, incr_num=1):

        res = await self.incrby(incr_num)
        if res > self._limit:
            return True
        return False


class RedisPoolManager(SafeSingleton, BasePoolManager):
    """Redis连接池管理器"""
    def __init__(self):
        super().__init__(RedisAsyncContextManager)

    def __await__(self):

        if not self._pools:

            for pool_name, pool_config in self._config.items():

                self._pools[pool_name] = yield from aioredis.create_pool(**pool_config).__await__()

                setattr(self._pools[pool_name], 'pool_name', pool_name)

                Utils.log.success(f'redis pool {self.pool_status(pool_name)} initialized')

            self._ready = True

        return self

    def _echo_pool_info(self, pool_name='default'):

        assert self._ready, f'{self.__class__.__name__} does not initialize'

        pool = self._pools[pool_name]

        if (pool.maxsize - pool.size + pool.freesize) < Utils.math.ceil(pool.maxsize / 3):
            Utils.log.warning(
                f'Redis pool not enough ({pool_name}): {pool.freesize}({pool.size}/{pool.maxsize})'
             )

    def close(self):
        if not self._pools:
            return
        pools, self._pools = self._pools, {}
        for pool in pools.values():
            pool.close()

    def pool_status(self, pool_name='default'):
        pool = self._pools[pool_name]
        return f'<[{pool_name}] {pool.address} [db:{pool.db}, size:[{pool.minsize}:{pool.maxsize}], free:{pool.freesize}]>'

    def allocate_lock(self, key, expire=60, pool_name='default'):

        return MutexLock(self._pools[pool_name], key, expire=expire)

    def allocate_limiter(self, pool_name='default', unique_tag=None, granularity=1, limit=10, ntp=None):
        return FrequencyLimiter(
            self._pools[pool_name],
            unique_tag=unique_tag,
            granularity=granularity,
            limit=limit, ntp=ntp
        )

    def func_partial(self, conn, pool_name):
        client = Redis(conn)
        self._contexts[pool_name].set(weakref.ref(client))
        return client


redis_pool_manager = RedisPoolManager()




from fastapi import HTTPException
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from starlette.requests import Request

from wintersweet.framework.response import Response, ErrResponse, HTTP400Response, HTTP500Response
from wintersweet.utils.base import Utils


async def http_exception_handler(
        request: Request, exc: HTTPException
) -> Response:

    headers = getattr(exc, r'headers', None)

    Utils.log.error(exc.detail)

    if headers:
        return HTTP500Response(headers=headers)
    else:
        return HTTP500Response()


async def request_validation_exception_handler(
        request: Request, exc: RequestValidationError
) -> Response:

    return HTTP400Response(content=jsonable_encoder(exc.errors()))




import aiomysql
from aioredis import Redis


class MysqlAsyncContextManager:
    __slots__ = ('_pool', '_commands_partial', '_sa_conn')

    def __init__(self, pool_or_conn, commands_partial):
        self._pool = self._sa_conn = None
        if not isinstance(pool_or_conn, aiomysql.Pool):
            self._sa_conn = pool_or_conn
        else:
            self._pool = pool_or_conn

        self._commands_partial = commands_partial

    async def __aenter__(self):
        if self._sa_conn is not None:
            return self._sa_conn
        conn = await self._pool.acquire()
        pool_name = getattr(self._pool, 'pool_name')
        self._sa_conn = self._commands_partial(conn, pool_name)
        return self._sa_conn

    async def __aexit__(self, exc_type, exc_value, tb):
        if self._pool is None:
            return
        await self._sa_conn.close()
        self._pool = None
        self._sa_conn = None


class RedisAsyncContextManager:
    __slots__ = ('_pool', '_sa_conn', '_commands_partial', 'conn')

    def __init__(self, pool_or_conn, commands_partial):
        self._sa_conn = self._pool = None

        if isinstance(pool_or_conn, Redis):
            self._sa_conn = pool_or_conn
        else:
            self._pool = pool_or_conn
        self._commands_partial = commands_partial

    async def __aenter__(self):
        if not self._sa_conn:
            conn = await self._pool.acquire()
            pool_name = getattr(self._pool, 'pool_name')
            self._sa_conn = self._commands_partial(conn, pool_name)

        return self._sa_conn

    async def __aexit__(self, exc_type, exc_value, tb):
        if self._pool is None:
            return
        self._sa_conn.close()

        await self._sa_conn.wait_closed()
        try:
            self._pool.release(self._sa_conn._pool_or_conn)
        finally:
            self._pool = None
            self._sa_conn = None


